import ast


def calculate_style_guide_adherence(parsed_code):
    # Calculates the adherence to a style guide of the parsed code.
    # Very dummy style_guide. Needs to be replaced by PEP8 Style
    style_guide = {
        ast.FunctionDef: {
            "args": {
                "vararg": None,
                "kwonlyargs": [],
                "kw_defaults": [],
                "kwarg": None
            }
        },
        ast.ClassDef: {"bases": [], "keywords": []},
        ast.Call: {"func": {"id": "print", "attr": None}},
        ast.Compare: {
            "ops": [
                ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE,
                ast.Is, ast.IsNot, ast.In, ast.NotIn
            ]
        },
        ast.Attribute: {"attr": None},
    }
    # Count the number of style guide violations in the parsed code
    style_violations = 0
    for node in ast.walk(parsed_code):
        if isinstance(node, tuple(style_guide.keys())):
            for attribute, allowed_values in style_guide[type(node)].items():
                if hasattr(node, attribute):
                    value = getattr(node, attribute)
                    if allowed_values is None:
                        style_violations += 1
                    else:
                        if isinstance(value, (list, tuple)):
                            for v in value:
                                if v not in allowed_values:
                                    style_violations += 1
                        elif value not in allowed_values:
                            style_violations += 1
    # Calculate the style guide adherence
    # as a fraction of the total number of nodes
    total_nodes = sum(1 for node in ast.walk(parsed_code))
    adherence = 1 - style_violations / total_nodes
    return adherence



import ast
from .comments import analyze_comment_quality
from .indentation import calculate_indentation_consistency
from .style_guide import calculate_style_guide_adherence
from .patterns import detect_repetitive_patterns
from .complexity import calculate_code_complexity


def extract_features(file_path, parsed_code):
    # Dict with features
    features = {}
    # Extract comment features
    features["comment_quality"] = analyze_comment_quality(
                                        file_path, parsed_code)
    features["code_identation"] = calculate_indentation_consistency(
                                        parsed_code)
    features["style_adherence"] = calculate_style_guide_adherence(
                                        parsed_code)
    features["repetitive_patterns"] = detect_repetitive_patterns(
                                        parsed_code)
    features["code_complexity"] = calculate_code_complexity(
                                        parsed_code)
    return features



import ast
import math
import numpy as np


def calculate_indentation_consistency(parsed_code, indentation_size=4):
    # Calculates the indentation consistency of the parsed code.
    indentation_levels = []
    # Traverse the AST and collect the indentation levels of all nodes
    for node in ast.walk(parsed_code):
        if isinstance(node, ast.TypeIgnore):
            # Exclude TypeIgnore nodes from the indentation consistency calculation
            continue
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            # Exclude function and class definitions
            # from the indentation consistency calculation
            continue
        if not isinstance(node, ast.AST):
            # Exclude non-AST nodes (such as comments)
            continue
        if not hasattr(node, "lineno"):
            # Exclude nodes without a line number (such as the root Module node)
            continue
        indentation_level = node.col_offset // indentation_size
        indentation_levels.append(indentation_level)
    # Calculate the median absolute deviation (MAD) of the indentation levels
    if len(indentation_levels) <= 1:
        # If there is only one or zero indentation levels,
        # return perfect consistency
        return 1.0
    else:
        mad = np.median(np.abs(indentation_levels - np.median(indentation_levels)))
        if (np.median(indentation_levels) <= 0):
            return 1.0
        else:
            return 1.0 - (mad / (np.median(indentation_levels) * 2))



import ast
from collections import defaultdict


def detect_repetitive_patterns(parsed_code):
    pattern_counts = defaultdict(int)
    # Traverse the AST and count the number of occurrences of each pattern
    for node in ast.walk(parsed_code):
        if hasattr(node, "lineno") and hasattr(node, "col_offset"):
            # Ignore nodes without line and column information (e.g., module)
            node_pattern = get_node_pattern(node)
            pattern_counts[node_pattern] += 1
    repetitive_patterns = [
        pattern for pattern, count in pattern_counts.items() if count > 1
    ]
    total_patterns = len(repetitive_patterns)
    num_nodes = sum(1 for _ in ast.walk(parsed_code))
    if num_nodes == 0:
        return 0.0
    else:
        return total_patterns / num_nodes


def get_node_pattern(node):
    if isinstance(node, ast.Call):
        # For function calls, include the function name and argument types
        arg_types = [get_node_pattern(arg) for arg in node.args]
        return f"Call:{get_node_pattern(node.func)}.{','.join(arg_types)}"
    elif isinstance(node, ast.Assign):
        # For assignments, include the target names and value type
        target_names = [get_node_pattern(target) for target in node.targets]
        value_type = get_node_pattern(node.value)
        return f"Assign:{','.join(target_names)}={value_type}"
    elif isinstance(node, ast.Name):
        # For variable names, return the name
        return node.id
    elif isinstance(node, ast.Attribute):
        # For attributes, recursively get the pattern of the value
        # and concatenate the attribute name
        return f"{get_node_pattern(node.value)}.{node.attr}"
    else:
        # For other node types, return the class name
        return node.__class__.__name__



import ast
import radon.complexity as radon_cc


def calculate_code_complexity(parsed_code):
    # Calculate the cyclomatic complexity
    cc = radon_cc.cc_visit_ast(parsed_code)
    total_complexity = sum(func.complexity for func in cc)
    return total_complexity



import pickle
from sklearn.tree import DecisionTreeClassifier


def train_model(pkl_path, labels, samples):
    # Train a decision tree classifier on the dataset
    clf = DecisionTreeClassifier()
    clf.fit(samples, labels)
    # Save the trained model to a file
    with open(pkl_path, "wb") as f:
        pickle.dump(clf, f)


def load(pkl_path):
    # Load the trained model from the pickle file
    with open(pkl_path, "rb") as f:
        clf = pickle.load(f)
    return clf


def predict(pkl_path, features):
    # Load the model
    model = load(pkl_path)
    # Convert the features into a feature vector
    feature_vector = [[features['comment_quality'],
                      features['code_identation'],
                      features['style_adherence'],
                      features['repetitive_patterns'],
                      features['code_complexity']]]
    # Use the trained model to make a prediction
    prediction = model.predict(feature_vector)
    # Return the prediction
    return bool(prediction[0])



import os
import ast
import pickle
import argparse

from botsniffer.feature_extraction.feature_extraction import extract_features
from botsniffer.ml_model.model import train_model, predict


def get_features(file_path, tree):
    features = extract_features(file_path, tree)
    feature_values = []
    for key, value in features.items():
        if isinstance(value, str):
            feature_values.append(float(value))
        elif isinstance(value, list):
            feature_values.extend(value)
        else:
            feature_values.append(float(value))
    sample = feature_values
    # For training, is expected that AI generated
    # files had "ai" in their name/path
    if "ai" in file_path.lower():
        label = True
    else:
        label = False
    return features, sample, label


def scan_path(path, identify, train):
    is_ai = False
    samples = []
    labels = []
    pkl_dir = os.path.join(os.path.dirname(__file__), 'data')
    pkl_path = os.path.join(
                        pkl_dir,
                        'botcode.pkl')
    if not os.path.exists(path):
        print("Error: '{}' does not exist.".format(path))
    elif os.path.isdir(path):
        # Run the feature extraction in a folder
        for root, dirs, files in os.walk(path):
            for file_name in files:
                if file_name.endswith(".py"):
                    file_path = os.path.join(root, file_name)
                    with open(file_path) as f:
                        tree = ast.parse(f.read(), type_comments=True)
                        features, sample, label = get_features(file_path, tree)
                        if identify:
                            is_ai = predict(pkl_path, features)
                            print('File:', file_path)
                            print('AI:', is_ai)
                            print('Features:', features)
                        elif train:
                            samples.append(sample)
                            if label:
                                labels.append(1)
                            else:
                                labels.append(0)
    else:
        if path.endswith(".py"):
            # Run the feature extraction in a single file
            with open(path) as f:
                tree = ast.parse(f.read(), type_comments=True)
                features, sample, label = get_features(path, tree)
                if identify:
                    is_ai = predict(pkl_path, features)
                    print('File:', path)
                    print('AI:', is_ai)
                    print('Features:', features)
                elif train:
                    samples.append(sample)
                    if label:
                        labels.append(1)
                    else:
                        labels.append(0)
        else:
            print("Error: '{}' is not a Python source code file.".format(path))
    if train:
        train_model(pkl_path, labels, samples)
        print('Trained:')
        print(' - labels:', len(labels))
        print(' - samples:', len(samples))
    elif identify:
        print('Done!')
    else:
        print("Error: you must specify either \
        --identify or --train.")


def main():
    parser = argparse.ArgumentParser(
        description="Scan Python source code files \
        and extract information about functions.")
    parser.add_argument("path", help="The path to scan.")
    parser.add_argument(
        "--identify",
        action="store_true",
        help="Identify AI-generated code.")
    parser.add_argument(
        "--train",
        action="store_true",
        help="Train a model to identify AI-generated code.")
    args = parser.parse_args()

    scan_path(args.path, args.identify, args.train)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import pathlib
import sys
from setuptools import setup, find_packages


HERE = pathlib.Path(__file__).parent
VERSION = '0.1'
PACKAGE_NAME = 'botsniffer'
AUTHOR = 'Oscar Valenzuela B.'
AUTHOR_EMAIL = 'alkamod@gmail.com'
URL = 'https://github.com/oscarvalenzuelab/botsniffer'
LICENSE = 'Apache-2.0'
DESCRIPTION = 'Detects AI generated code using ML'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    packages=[
        'botsniffer',
        'botsniffer.data',
        'botsniffer.feature_extraction',
        'botsniffer.ml_model'],
    entry_points={"console_scripts": ["botsniffer=botsniffer.scanner:main"]},
    install_requires=["sklearn", "numpy", "pandas", "radon"],
    url=URL,
    package_data={
        "botsniffer.data": ["*.txt", "*.ini", "*.pkl"],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License'
    ],
)



import ast
import tokenize
from io import BytesIO


def analyze_comment_quality(file_path, parsed_code):
    # This code tokenize the content of the file
    # looking for comments. I planned to use AST
    # but after python 3.8 comments output it's
    # not supported also, I'm tired and out of
    # coffee, so implementation is dummy it works
    total_lines = 0
    total_comments = 0
    high_quality_comments = 0
    fObj = open(file_path, 'r')
    for toktype, tok, start, end, line in (
        tokenize.generate_tokens(fObj.readline)
    ):
        total_lines += 1
        if toktype == tokenize.COMMENT:
            total_comments += 1
            if is_high_quality_comment(tok):
                high_quality_comments += 1
    if high_quality_comments > 0:
        quality = (high_quality_comments / total_comments) + 1
    elif total_comments > 0:
        quality = total_comments / total_lines
    else:
        quality = 0.0
    return quality


def is_high_quality_comment(comment):
    # Set of words and phrases that indicate a high-quality comment
    high_quality_words = {"TODO", "FIXME", "NOTE", "HACK", "FIX", "DEFINE"}
    # Check if the comment contains any high-quality words or phrases
    for word in high_quality_words:
        if word in comment.upper():
            return True
    return False
